// Copyright (C) 2010 - DIGITEO - Michael Baudin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

function [x,fval,exitflag,output,lambda,grad,hessian] = fmincon_ipopt ( fun,x0,A,b,Aeq,beq,lb,ub,nonlcon,options )
  // Uses IPOPT as an implementation of fmincon.
  //
  // Calling Sequence
  //   x = fmincon_ipopt(fun,x0,A,b)
  //   x = fmincon_ipopt(fun,x0,A,b,Aeq,beq)
  //   x = fmincon_ipopt(fun,x0,A,b,Aeq,beq,lb,ub)
  //   x = fmincon_ipopt(fun,x0,A,b,Aeq,beq,lb,ub,nonlcon)
  //   x = fmincon_ipopt(fun,x0,A,b,Aeq,beq,lb,ub,nonlcon,options)
  //   [x,fval,exitflag,output,lambda,grad,hessian] = fmincon_ipopt ( ... )
  //
  // Parameters
  //   fun: a function, the function to minimize. See below for the complete specifications.
  //   x0: a nx1 or 1xn matrix of doubles, where n is the number of variables. The initial guess for the optimization algorithm.
  //   A: a nil x n matrix of doubles, where n is the number of variables and nil is the number of linear inequalities.
  //   b: a nil x 1 matrix of doubles, where nil is the number of linear inequalities.
  //   Aeq: a nel x n matrix of doubles, where n is the number of variables and nel is the number of linear equalities.
  //   beq: a nel x 1 matrix of doubles, where nel is the number of linear inequalities.
  //   lb: a nx1 or 1xn matrix of doubles, where n is the number of variables. The lower bound for x.
  //   ub: a nx1 or 1xn matrix of doubles, where n is the number of variables. The upper bound for x.
  //   nonlcon: a function, the nonlinear constraints. See below for the complete specifications.
  //   options: an optional struct, as provided by optimset
  //
  // Description
  //   We assume that the arguments of fmincon_ipopt have already been checked by fmincon.
  //   This function is not designed to be available directly to the end-user.
  //
  // Authors
  // Michael Baudin, DIGITEO, 2010

  //===================================================
  // Extract options from data structure
  defaultoptions = optimset ("fmincon");
  MaxFunEvals = optimget ( options , "MaxFunEvals" , 3000 );
  MaxIter = optimget ( options , "MaxIter" , 1000 );
  TolFun = optimget ( options , "TolFun" , defaultoptions.TolFun );
  TolX = optimget ( options , "TolX" , 1.e-10 );
  Display = optimget ( options , "Display" , defaultoptions.Display );
  OutputFcn = optimget ( options , "OutputFcn" , defaultoptions.OutputFcn );
  PlotFcns = optimget ( options , "PlotFcns" , defaultoptions.PlotFcns );
  TolCon = optimget ( options , "TolCon" , defaultoptions.TolCon );
  //===================================================
  // Configure options of ipopt from user's options
  // Compute the number of constraints
  nbvar = size(x0,"*")
  nA = size(A,"r")
  nAeq = size(Aeq,"r")
  if ( nonlcon == [] ) then
    nc = 0
    nceq = 0
  else
    [c, ceq] = nonlcon ( x0 )
    if ( c == [] ) then
      nc = 0
    else
      nc = size(c,"r")
      if ( size(c,"c") <> 1 ) then
        errmsg = msprintf(gettext("%s: Wrong number of columns for the variable c at output of nonlcon: expected 1, but got %d instead."),"fmincon", size(c,"c") );
        error(errmsg);
      end
    end
    if ( ceq == [] ) then
      nceq = 0
    else
      nceq = size(ceq,"r")
      if ( size(ceq,"c") <> 1 ) then
        errmsg = msprintf(gettext("%s: Wrong number of columns for the variable ceq at output of nonlcon: expected 1, but got %d instead."),"fmincon", size(ceq,"c") );
        error(errmsg);
      end
    end
  end
  nb_constr = nc + nceq + nA + nAeq
  // The sparsity structure of the constraints
  for ic = 1 : nb_constr
    for iv = 1 : nbvar
      k = iv + (ic-1)*nbvar
      sparse_dg(k,1:2) = [ic iv];
    end
  end
  // The sparsity structure of the Lagrangian
  for i = 1 : nbvar
    for j = 1 : nbvar
      k = j + (i-1)*nbvar
      sparse_dh(k,1:2) = [i j];
    end
  end
  // Store the rhs/lhs of the constraints
  k = 1
  // Store rhs/lhs of the nonlinear inequality constraints
  constr_lhs(k:k+nc-1) = - %inf * ones(nc,1)
  constr_rhs(k:k+nc-1) = zeros(nc,1)
  k = k + nc
  // Store rhs/lhs of the nonlinear equality constraints
  constr_lhs(k:k+nceq-1) = zeros(nceq,1)
  constr_rhs(k:k+nceq-1) = zeros(nceq,1)
  k = k + nceq
  // Store rhs/lhs of the linear inequality constraints
  constr_lhs(k:k+nA-1) = zeros(nA,1)
  constr_rhs(k:k+nA-1) = b
  k = k + nA
  // Store rhs/lhs of the linear inequality constraints
  constr_lhs(k:k+nAeq-1) = zeros(nAeq,1)
  constr_rhs(k:k+nAeq-1) = beq
  k = k + nAeq
  // Set type of each variable: 
  // I do not have the necessary information from the input arguments
  // I set all variable to nonlinear = 1
  var_lin_type = ones(nbvar,1)
  // Set the type of each constraint
  // nonlinear = 1
  // linear = 0
  constr_lin_type = []
  k = 1
  // Store type of the nonlinear inequality constraints
  constr_lin_type(k:k+nc-1) = ones(nc,1)
  k = k + nc
  // Store type of the nonlinear equality constraints
  constr_lin_type(k:k+nceq-1) = ones(nceq,1)
  k = k + nceq
  // Store type of the linear inequality constraints
  constr_lin_type(k:k+nA-1) = zeros(nA,1)
  k = k + nA
  // Store type of the linear inequality constraints
  constr_lin_type(k:k+nAeq-1) = zeros(nAeq,1)
  k = k + nAeq
// Setup default bounds
  if ( lb == [] ) then
    lb = -%inf * ones(nbvar,1)
  end
  if ( ub == [] ) then
    ub = %inf * ones(nbvar,1)
  end
  //
  // Configure special settings for IPOPT
  params = init_param();
  //
  // Configure the computation of the Hessian matrix.
  params = add_param(params,"hessian_approximation","limited-memory");
  ipopt_hessianLagr = []
  //params = add_param(params,"hessian_approximation","exact");
  if ( %f ) then
    // TUNE FOR DEBUG ONLY !!!!!!!!!
    // We use the given Hessian
    //params = add_param(params,"hessian_approximation","exact");
    // We use a limited-bfgs approximation for the Hessian.
    //params = add_param(params,"hessian_approximation","limited-memory");
    params = add_param(params,"print_level",12);
    params = add_param(params,"journal_level",12);
  end
  //===================================================
  // Store data in a global variable
  global ipopt_data
  ipopt_data = tlist(["TIPOPT"
    "nbvar"
    "nc"
    "nceq"
    "nA"
    "nAeq"
    "fun" 
    "x0"
    "A"
    "b"
    "Aeq"
    "beq"
    "lb"
    "ub"
    "nonlcon"
    "verbose"
    "options"
    "lastXObj"
    "storeObj"
    "storeGradObj"
    "lastXCon"
    "storeCon"
    "storeGradCon"
    "storeConEq"
    "storeGradConEq"
    ]);
  ipopt_data.nbvar = nbvar
  ipopt_data.nc = nc
  ipopt_data.nceq = nceq
  ipopt_data.nA = nA
  ipopt_data.nAeq = nAeq
  ipopt_data.fun = fun
  ipopt_data.x0 = x0
  ipopt_data.A = A
  ipopt_data.b = b
  ipopt_data.Aeq = Aeq
  ipopt_data.beq = beq
  ipopt_data.lb = lb
  ipopt_data.ub = ub
  ipopt_data.nonlcon = nonlcon
  //ipopt_data.verbose = 1
  ipopt_data.verbose = 0
  ipopt_data.options = options
  //
  ipopt_data.lastXObj = []
  ipopt_data.storeObj = []
  ipopt_data.storeGradObj = []
  //
  ipopt_data.lastXCon = []
  ipopt_data.storeCon = []
  ipopt_data.storeConEq = []
  ipopt_data.storeGradCon = []
  ipopt_data.storeGradConEq = []

  //===================================================
  // Call Ipopt
  [x fval extra] = ipopt ( x0, ipopt_f, ipopt_df, ..
    ipopt_g, ipopt_dg, sparse_dg, ipopt_hessianLagr, sparse_dh, ..
    var_lin_type, constr_lin_type, constr_rhs, constr_lhs, lb, ub , params );
  //===================================================
  // Set output arguments
  //
  // Set exitflag
  // ipopt2fmincon : map from ipopt status to fmincon status
  ipopt2fmincon = [
    0 1
    1 1
    2 -2
    3 %nan
    4 %nan
    5 -1
    6 %nan
    -1 0
    -2 %nan
    -3 %nan
    -10 %nan
    -11 %nan
    -12 %nan
    -13 %nan
    -100 %nan 
    -101 %nan
    -102 %nan
    -199 %nan
  ];
  // 0 : Number of iterations exceeded options.MaxIter or number of function evaluations exceeded options.FunEvals.
  // 1 : First-order optimality measure was less than options.TolFun, and maximum constraint violation was less than options.TolCon.
  // 2 : Change in x was less than options.TolX and maximum constraint violation was less than options.TolCon.
  // -1 : The output function terminated the algorithm.
  // -2 : No feasible point was found.
  // -3 : Current point x went below options.ObjectiveLimit and maximum constraint violation was less than options.TolCon.
  // %nan : other problem
  // TODO : fmincon = 2 or -3

  ipopt_status = extra('status')
  ist = find ( ipopt2fmincon(:,1) == ipopt_status )
  if ( ist <> [] ) then
    // "by chance", the status of ipopt is known
    exitflag = ipopt2fmincon(ist,2)
  else
    exitflag = %nan
  end
  //
  // Set grad
  grad = ipopt_df ( x )'
  //
  // Set hessian
  hessian = ipopt_hessian ( x )
  //
  // Set lambda
  lambda = struct(...
      "lower"      , [], ..
      "upper"      , [], ..
      "eqlin"      , [], ..
      "eqnonlin"   , [], ..
      "ineqlin"    , [], ..
      "ineqnonlin" , []);
  iplam = extra("lambda")
  lambda.lower = zeros(nbvar,1)
  lambda.upper = zeros(nbvar,1)
  k = 1
  if ( nc == 0 ) then
    lambda.ineqnonlin = []
  else
    lambda.ineqnonlin = iplam(k:k+nc-1)
    k = k + nc
  end
  if ( nceq == 0 ) then
    lambda.eqnonlin = []
  else
    lambda.eqnonlin = iplam(k:k+nceq-1)
    k = k + nceq
  end
  if ( nA == 0 ) then
    lambda.ineqlin = []
  else
    lambda.ineqlin = iplam(k:k+nA-1)
    k = k + nA
  end
  if ( nAeq == 0 ) then
    lambda.eqlin = []
  else
    lambda.eqlin = iplam(k:k+nAeq-1)
    k = k + nAeq
  end
  // Set output
  output = struct(..
      "iterations"      , [], ..
      "funcCount"       , [], ..
      "stepsize"        , [], ..
      "algorithm"       , [], ..
      "firstorderopt"   , [], ..
      "constrviolation" , [], ..
      "cgiterations" , [], ..
      "message"         , [] );
  output.iterations = extra('it_count')
  output.funcCount = extra('fobj_eval')
  output.stepsize = []
  output.algorithm = "ipopt"
  output.firstorderopt = extra('kkt_error')
  output.constrviolation = extra('constr_viol')
  output.cgiterations = 0
  output.message = sprintf("%s %s\n%s\n%s", "Optimization terminated:",..
    "first-order optimality relative error", ..
    "less than options.TolFun, and relative constraint violation less", ..
    "than options.TolCon.");
  //
  // Set x
  x = x'
endfunction


//
// ipopt_f --
//   Evaluates the objective function.
function y = ipopt_f ( x )
  global ipopt_data
  //
  // Extract fields
  fun = ipopt_data.fun
  //
  if ( ipopt_data.verbose == 1 ) then
    mprintf ("ipopt_f\n")
  end
//pause
  GradObj = optimget ( ipopt_data.options , "GradObj" , "off" );
  if ( GradObj == "off" ) then
    y = fun ( x )
  else
    if ( ipopt_data.lastXObj == x ) then
      y = ipopt_data.storeObj
    else
      [f,df] = fun ( x )
      ipopt_data.lastXObj = x
      ipopt_data.storeObj = f
      ipopt_data.storeGradObj = df
      y = f
    end
  end
endfunction
//
// ipopt_df --
//   Evaluates the gradient of the objective function.
function y = ipopt_df ( x )
  global ipopt_data
  //
  // Extract fields
  fun = ipopt_data.fun
  //
  if ( ipopt_data.verbose == 1 ) then
    mprintf ("ipopt_df\n")
  end
//pause
  GradObj = optimget ( ipopt_data.options , "GradObj" , "off" );
  if ( GradObj == "off" ) then
    y = derivative ( fun , x )
  else
    if ( ipopt_data.lastXObj == x ) then
      y = ipopt_data.storeGradObj
    else
      [f,df] = fun ( x )
      ipopt_data.lastXObj = x
      ipopt_data.storeObj = f
      ipopt_data.storeGradObj = df
      y = df
    end
  end
endfunction
//
// ipopt_g --
//   Evaluates the inequality constraints.
function y = ipopt_g ( x )
  global ipopt_data
//pause
  //
  // Extract fields
  nonlcon = ipopt_data.nonlcon
  A = ipopt_data.A
  b = ipopt_data.b
  Aeq = ipopt_data.Aeq
  beq = ipopt_data.beq
  nc = ipopt_data.nc
  nceq = ipopt_data.nceq
  nA = ipopt_data.nA
  nAeq = ipopt_data.nAeq
  //
  if ( ipopt_data.verbose == 1 ) then
    mprintf ("ipopt_g\n")
  end
  //
  // Compute non linear constraints
  [c , ceq] = nonlcon ( x )
  //
  k = 1
  // Store nonlinear inequality constraints
  y(k:k+nc-1) = c
  k = k + nc
  // Store nonlinear equality constraints
  y(k:k+nceq-1) = ceq
  k = k + nceq
  // Store linear inequality constraints
  y(k:k+nA-1) = A * x
  k = k + nA
  // Store linear equality constraints
  y(k:k+nAeq-1) = Aeq * x
endfunction

//
// ipopt_nonlcon --
//   Returns a column vector made of the non linear 
//   inequality and equality constraints.
//   This function is used by derivative, when the derivatives
//   are not given by the user.
function y = ipopt_nonlcon ( x )
  global ipopt_data
//pause
  //
  // Extract fields
  nonlcon = ipopt_data.nonlcon
  nc = ipopt_data.nc
  nceq = ipopt_data.nceq
  nA = ipopt_data.nA
  nAeq = ipopt_data.nAeq
  //
  // Compute constraints
  [c , ceq] = nonlcon ( x )
  //
  k = 1
  // Store nonlinear inequality constraints
  y(k:k+nc-1) = c
  k = k + nc
  // Store nonlinear equality constraints
  y(k:k+nceq-1) = ceq
endfunction
//
// ipopt_dg --
//   Evaluates the gradient of the inequality constraints.
function y = ipopt_dg ( x )
  // Extract data from the problem
  global ipopt_data
//pause
  //
  // Extract fields
  nonlcon = ipopt_data.nonlcon
  nbvar = ipopt_data.nbvar
  nc = ipopt_data.nc
  nceq = ipopt_data.nceq
  nA = ipopt_data.nA
  nAeq = ipopt_data.nAeq
  A = ipopt_data.A
  Aeq = ipopt_data.Aeq
  //
  if ( ipopt_data.verbose == 1 ) then
    mprintf ("ipopt_dg\n")
  end
  //
  // Compute derivatives of nonlinear constraints : DC and DCeq
  GradConstr = optimget ( ipopt_data.options , "GradConstr" , "off" )
  if ( GradConstr == "off" ) then
    DCall = derivative ( ipopt_nonlcon , x )
    DC   ( 1 : nc   , 1:nbvar) = DCall ( 1    : nc      , 1:nbvar)
    DCeq ( 1 : nceq , 1:nbvar) = DCall ( nc+1 : nc+nceq , 1:nbvar)
  else
    if ( ipopt_data.lastXCon == x ) then
      DC = ipopt_data.storeGradCon
      DCeq = ipopt_data.storeGradConEq
    else
      [c,ceq,DC,DCeq] = nonlcon ( x )
      // Transpose DC and DCeq : 
      // we want them row by row, but they are 
      // computed column by column
      DC = DC'
      DCeq = DCeq'
      ipopt_data.lastXCon = x
      ipopt_data.storeCon = c
      ipopt_data.storeConEq = ceq
      ipopt_data.storeGradCon = DC
      ipopt_data.storeGradConEq = DCeq
    end
  end
  //
  // Merge DC, DCeq, A and Aeq into the derivative
  k = 1
  // Store gradient of nonlinear inequality constraints
  y(k:k+nc*nbvar-1) = matrix(DC',nc*nbvar,1)
  k = k + nc*nbvar
  // Store gradient of nonlinear equality constraints
  y(k:k+nceq*nbvar-1) = matrix(DCeq',nceq*nbvar,1)
  k = k + nceq*nbvar
  // Store gradient of linear inequality constraints
  y(k:k+nA*nbvar-1) = matrix(A',nA*nbvar,1)
  k = k + nA*nbvar
  // Store gradient of linear equality constraints
  y(k:k+nAeq*nbvar-1) = matrix(Aeq',nAeq*nbvar,1)
endfunction
//
// ipopt_hessianLagr --
//   Evaluates the Hessian of the Lagrangian
// TODO : take into account for ceq
// TODO : generalize to arbitrary number of constraints !!!
// function HL = ipopt_hessianLagr ( x , lambda , obj_weight )
//   global ipopt_data
//   //pause
//   //
//   // Extract fields
//   nbvar = ipopt_data.nbvar
//   nc = ipopt_data.nc
//   nceq = ipopt_data.nceq
//   nA = ipopt_data.nA
//   nAeq = ipopt_data.nAeq
//   fun = ipopt_data.fun
//   if ( ipopt_data.verbose == 1 ) then
//     mprintf ("ipopt_hessianLagr\n")
//   end
//   [Jf , Hf] = derivative(fun,x,%eps^(1/3),1,"blockmat" )
//   HL = obj_weight * Hf
//   [Jg1 , Hg1] = derivative(list(ipopt_gi,1),x,%eps^(1/3),1,"blockmat" )
//   HL = HL + lambda(1) * Hg1
//   [Jg2 , Hg2] = derivative(list(ipopt_gi,2),x,%eps^(1/3),1,"blockmat" )
//   HL = HL + lambda(2) * Hg2
// endfunction
//
// ipopt_gi --
//   Evaluates the i-th component of the constraints
function y = ipopt_gi ( x , i )
  g = ipopt_dg ( x )
  y = g(i)
endfunction

// ipopt_hessian --
//   Evaluates the Hessian of the objective function
function y = ipopt_hessian ( x )
  global ipopt_data
//pause
  fun = ipopt_data.fun
  if ( ipopt_data.verbose == 1 ) then
    mprintf ("ipopt_hessian\n")
  end
  [Jf , Hf] = derivative(fun,x,%eps^(1/3),1,"blockmat" )
  y = Hf
endfunction



